﻿using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using VRageMath;
using UIFun;
using VRage.Game;
using Sandbox.Graphics;
using VRage.Utils;

namespace UIFun.Messages
{
	class AdvHUDMessage : MessageBase
	{
		public Vector2D Origin;
		Options options;
		Vector4 Shadow;

		public AdvHUDMessage()
		{
			id = 0;
			ttl = 0;
			Origin = Vector2D.Zero;
			message = "";
			Color shadowcolor = Color.Black;
			shadowcolor.A = 1;
			Shadow = shadowcolor.ToVector4();
			options = Options.None;
		}
		public AdvHUDMessage(long mid, int timetolive, Vector2D origin, double scale, bool hidehud, bool shadowing, Vector4 shadow, string message)
		{
			options = Options.None;
			this.scale = scale;
			id = mid;
			ttl = timetolive;
			Origin = origin;
			this.Shadow = shadow;
			if (hidehud)
				options |= Options.HideHud;
			if (shadowing)
				options |= Options.Shadowing;
			this.message = message;
		}

		public override long Decode(byte[] obj)
		{
			try
			{
				int index = 0;

				ushort modver = BitConverter.ToUInt16(obj, index);
				ushort msgtype = BitConverter.ToUInt16(obj, index += sizeof(short));
				if (msgtype != 3) return -1;//fail
				long modid = BitConverter.ToInt64(obj, index += sizeof(short));
				id = BitConverter.ToInt64(obj, index += sizeof(long));
				ttl = BitConverter.ToInt32(obj, index += sizeof(long));
				double vx = BitConverter.ToDouble(obj, index += sizeof(int));
				double vy = BitConverter.ToDouble(obj, index += sizeof(double));
				Origin = new Vector2D(vx, vy);
				options = (Options)obj[index += sizeof(double)];
				float cx = BitConverter.ToSingle(obj, index += 1);
				float cy = BitConverter.ToSingle(obj, index += sizeof(float));
				float cz = BitConverter.ToSingle(obj, index += sizeof(float));
				float cw = BitConverter.ToSingle(obj, index += sizeof(float));
				scale = BitConverter.ToDouble(obj, index += sizeof(float));
				Shadow = new Vector4(cx, cy, cz, cw);
				index += sizeof(double);
				message = Encoding.UTF8.GetString(obj, index, obj.Length - index);

				return modid;
			}
			catch
			{
				ttl = -1;
				return -1;
			}

		}



		public override void Write()
		{
			try
			{
				if(options.HasFlag(Options.HideHud))
				{
					if (MyAPIGateway.Session.Config.MinimalHud)
						return;//hud is hidden dont draw. 
				}
				Vector2D origin = Origin;
				if (MyAPIGateway.Session?.Camera == null) return;
				Regex ColorEncoder = new Regex("<color=(\\w*?|\\d*?\\,\\d*?\\,\\d*?|\\d*?,\\d*?,\\d*?,\\d*?)>", RegexOptions.IgnoreCase);
				Vector4 textcolor = Color.White.ToVector4();
				var matches = ColorEncoder.Split(message);
				double aspectratio = (double)MyAPIGateway.Session.Config.ScreenWidth / (double)MyAPIGateway.Session.Config.ScreenHeight;
				//double aspectratio = MyAPIGateway.Session.Camera.ViewportSize.X / MyAPIGateway.Session.Camera.ViewportSize.X;
                var inv_aspectratio = 1 / aspectratio;
				var fov = MyAPIGateway.Session.Camera.FovWithZoom;
				double localscale = 0.1735 * Math.Tan(fov / 2);//exact :)

				origin.X *= localscale * (aspectratio * 9d / 16d);//close enough
				origin.Y *= localscale * inv_aspectratio * (aspectratio * 9d / 16d);
			
				float size = 0.002f;
				float plusx = 0;
				float plusy = 0;
				int i = 0;
				foreach (var match in matches)
				{
					i++;
					if (i % 2 == 0)
					{
						Color foundColor;
						if (FontTexture.TryGetColor(match, out foundColor))
						{
							textcolor = foundColor.ToVector4();
						}
						continue;
					}
					foreach (var letter in match)
					{
						if (letter == '\n')
						{
							plusy -= (0.0025f * (float)(localscale * 9d * scale));
							plusx = 0;
							continue;
						}
						if (options.HasFlag(Options.Shadowing))
							WriteShadow(origin + new Vector2D(plusx, plusy), letter, Shadow, localscale * 9d * scale);
						size = WriteLetter(origin + new Vector2D(plusx, plusy), letter, textcolor, localscale * 9d * scale);
						plusx += size;
					}
				}

			}
			catch
			{

			}


		}
		private float WriteLetter(Vector2D Origin, char letter, Vector4 fontcolor, double scale)
		{
			FontTexture.FontMatrix value;

			MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X, Origin.Y, -0.1), headmatrix);
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(letter != ' ')
					VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor, textpos, headmatrix.Up, headmatrix.Left, 0.002f * (float)scale);
				return ((float)value.ex / (float)45) * 0.0025f * (float)scale;
			}
			return 0.002f * (float)scale;
		}
		private float WriteShadow(Vector2D Origin, char letter, Vector4 fontcolor, double scale)
		{

			FontTexture.FontMatrix value;
			MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X - 0.0005 * scale, Origin.Y, -0.10001), headmatrix);
			
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(letter != ' ')
					VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute("EngineThrustMiddle"), fontcolor, textpos, headmatrix.Up, headmatrix.Left, 0.0025f * (float)scale);
				return ((float)value.ex / (float)45) * 0.0025f * (float)scale;
			}
			return 0.002f * (float)scale;
		}
	}
}
