﻿using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using VRage.Game;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messages
{
	class BillBoardHUDMessage : MessageBase
	{
		public Vector2D Origin;
		Vector4 BillboardColor;
		Options options;
		float angle = 0;
		public BillBoardHUDMessage()
		{
			id = 0;
			ttl = 0;
			Origin = Vector2D.Zero;
			message = "";
			Color shadowcolor = Color.Black;
			shadowcolor.A = 1;
			BillboardColor = shadowcolor.ToVector4();
			options = Options.None;
			angle = 0f;
		}
		public BillBoardHUDMessage(long mid, int timetolive, Vector2D origin, double scale, float angle, bool hidehud, Vector4 shadow, string message)
		{
			options = Options.None;
			this.scale = scale;
			id = mid;
			ttl = timetolive;
			Origin = origin;
			this.BillboardColor = shadow;
			if (hidehud)
				options |= Options.HideHud;
			this.message = message;
			this.angle = angle;
		}

		public override long Decode(byte[] obj)
		{
			try
			{
				int index = 0;

				ushort modver = BitConverter.ToUInt16(obj, index);
				ushort msgtype = BitConverter.ToUInt16(obj, index += sizeof(short));
				if (msgtype != 4) return -1;//fail
				long modid = BitConverter.ToInt64(obj, index += sizeof(short));
				id = BitConverter.ToInt64(obj, index += sizeof(long));
				ttl = BitConverter.ToInt32(obj, index += sizeof(long));
				double vx = BitConverter.ToDouble(obj, index += sizeof(int));
				double vy = BitConverter.ToDouble(obj, index += sizeof(double));
				Origin = new Vector2D(vx, vy);
				options = (Options)obj[index += sizeof(double)];
				float cx = BitConverter.ToSingle(obj, index += 1);
				float cy = BitConverter.ToSingle(obj, index += sizeof(float));
				float cz = BitConverter.ToSingle(obj, index += sizeof(float));
				float cw = BitConverter.ToSingle(obj, index += sizeof(float));
				scale = BitConverter.ToDouble(obj, index += sizeof(float));
				angle = BitConverter.ToSingle(obj, index += sizeof(double));
				BillboardColor = new Vector4(cx, cy, cz, cw);
				index += sizeof(float);
				message = Encoding.UTF8.GetString(obj, index, obj.Length - index);
				return modid;
			}
			catch
			{
				ttl = -1;
				return -1;
			}

		}
		public override void Write()
		{
			try
			{
				if (options.HasFlag(Options.HideHud))
				{
					if (MyAPIGateway.Session.Config.MinimalHud)
						return;//hud is hidden dont draw. 
				}
				if (MyAPIGateway.Session?.Camera == null) return;
				Vector2D origin = Origin;
				double aspectratio = (double)MyAPIGateway.Session.Config.ScreenWidth / (double)MyAPIGateway.Session.Config.ScreenHeight;
				var inv_aspectratio = 1 / aspectratio;
				var fov = MyAPIGateway.Session.Camera.FovWithZoom;
				double localscale = 0.1735 * Math.Tan(fov / 2);//exact :)
				origin.X *= localscale * (aspectratio * 9d / 16d);//close enough
				origin.Y *= localscale * inv_aspectratio * (aspectratio * 9d / 16d);
				WriteBillboard(origin, message, BillboardColor, localscale * 9d * scale);
			}
			catch
			{

			}
		}
		private void WriteBillboard(Vector2D Origin, string billboard, Vector4 billboardcolor, double scale)
		{
			MatrixD cameramatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X, Origin.Y, -0.1), cameramatrix);
			var rotquad = Quaternion.CreateFromAxisAngle(cameramatrix.Forward, angle);
			cameramatrix = MatrixD.Transform(cameramatrix, rotquad);
			VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(billboard), billboardcolor, textpos, cameramatrix.Up, cameramatrix.Left, 0.002f * (float)scale);
			return;
		}
	}
}
