﻿using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using VRage.Game;
using VRage.Game.Entity;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messages
{
	class EntityMessage : MessageBase
	{
		public MyEntity Entity;
		public long EntityId;
		public Vector3D rel;
		public Vector3D up;
		public Vector3D forward;
		public MatrixD local;
		public FontTexture.TextOrientation orient;
		public Vector2D Max;
		//note add LOD. 
		public override long Decode(byte[] obj)
		{
			try
			{
				int index = 0;
				
				ushort modver = BitConverter.ToUInt16(obj, index);
				ushort msgtype = BitConverter.ToUInt16(obj, index += sizeof(short));
				if (msgtype != 2) return -1;//fail
				
				long modid = BitConverter.ToInt64(obj, index += sizeof(short));
				id = BitConverter.ToInt64(obj, index += sizeof(long));
				
				ttl = BitConverter.ToInt32(obj, index += sizeof(long));
				
				EntityId = BitConverter.ToInt64(obj, index += sizeof(int));
				if (!MyAPIGateway.Entities.EntityExists(EntityId))
					return -1;
				Entity = MyAPIGateway.Entities.GetEntityById(EntityId) as MyEntity;
				if (Entity == null) return -1;
				orient = (FontTexture.TextOrientation)obj[index += sizeof(long)];
				Max = new Vector2D(BitConverter.ToDouble(obj, index += 1), BitConverter.ToDouble(obj, index += sizeof(double))); 
				rel = new Vector3D(BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				up = new Vector3D(BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				forward = new Vector3D(BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				up.Normalize();
				forward.Normalize();
				scale = BitConverter.ToDouble(obj, index += sizeof(double));
				index += sizeof(double);
				message = Encoding.UTF8.GetString(obj, index, obj.Length - index);
				
				local = MatrixD.CreateWorld(rel, up, forward);
				return modid;
			}
			catch
			{
				ttl = -1;
				return -1;//fail
			}
		}
		public override void Write()
		{
			if (MyAPIGateway.Session?.Camera == null) return;
			Regex ColorEncoder = new Regex("<color=(\\w*?|\\d*?\\,\\d*?\\,\\d*?|\\d*?,\\d*?,\\d*?,\\d*?)>", RegexOptions.IgnoreCase);
			Color textcolor = Color.White;
			if(Entity == null)
			{
				ttl = -1;
				return;
			}
			try
			{

				if (Vector3.DistanceSquared(MyAPIGateway.Session.Camera.WorldMatrix.Translation, Entity.WorldMatrix.Translation) > Math.Pow(200d * (scale < 1.0d ? 1.0d : scale), 2))
					return;
				var matches = ColorEncoder.Split(message);



				float size = 0;
				float xstart = 0;
				if (orient == FontTexture.TextOrientation.center || orient == FontTexture.TextOrientation.rtl)
					xstart = -calculateStart(matches);
				if (orient == FontTexture.TextOrientation.rtl)
				{
					//matches.Reverse();
				}
				if (orient == FontTexture.TextOrientation.center)
				{
					xstart = xstart / 2f;
				}
				float plusx = xstart;
				float plusy = 0;
				int i = 0;
				foreach (var match in matches)
				{
					i++;
					if (i % 2 == 0)
					{
						Color foundColor;
						if (FontTexture.TryGetColor(match, out foundColor))
						{
							textcolor = foundColor;
						}
						continue;
					}
					foreach (var letter in match)
					{
						if (letter == '\n')
						{
							plusy -= (float)(scale) * 1.2f;
							plusx = xstart;
							continue;
						}

						MatrixD current = Entity.WorldMatrix;
						var rot = Quaternion.CreateFromRotationMatrix(local.GetOrientation());
						var rot2 = Quaternion.CreateFromRotationMatrix(current.GetOrientation());
						var rotby = Quaternion.Concatenate(rot, rot2);
						current = MatrixD.CreateFromQuaternion(rotby);
						current.Translation = Entity.WorldMatrix.Translation + Vector3D.Transform(local.Translation, current);

						Vector3D vec_x = new Vector3D(current.Left) * -plusx;
						Vector3D vec_y = new Vector3D(current.Up) * plusy;
						current.Translation = current.Translation + vec_x + vec_y;
						
						if(( plusx < Max.X || Max.X <=0 ) && (plusy > -Max.Y || Max.Y <= 0))
						{
							size = WriteLetter(current, letter, textcolor, scale);
							plusx += size;
						}

					}
				}

			}
			catch
			{
				return;
			}
		}
		private float calculateStart(string[] matches)
		{
			float cnt = 0.0f;
			float mxcnt = 0.0f;
			FontTexture.FontMatrix value;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						if (cnt > mxcnt)
							mxcnt = cnt;
						cnt = 0.0f;
						continue;
					}
					if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
					{
						cnt += ((float)value.ex / (float)45) * (float)scale * 1.2f;
					}
					else
						cnt += ((float)15 / (float)45) * (float)scale * 1.2f;

				}
			}
			if (cnt > mxcnt)
				mxcnt = cnt;

			return mxcnt;
		}
		private float WriteLetter(MatrixD world, char letter, Color fontcolor, double scale)
		{
			FontTexture.FontMatrix value;
			Vector3D textpos = world.Translation;
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor.ToVector4(), textpos, world.Up, world.Left, (float)scale);
				return ((float)value.ex / (float)45) * (float)scale * 1.2f;
			}
			return ((float)15 / (float)45) * (float)scale * 1.2f;
		}


	}
}
