﻿using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using VRageMath;
using UIFun;
using VRage.Game;
using VRage.Utils;

namespace UIFun.Messages
{
	class HUDMessage : MessageBase
	{
		public Vector2D Origin;

		public HUDMessage()
		{
			id = 0;
			ttl = 0;
			Origin = Vector2D.Zero;
			message = "";
		}
		public HUDMessage(long mid, int timetolive, Vector2D origin, string message)
		{
			scale = 1;
			id = mid;
			ttl = timetolive;
			Origin = origin;
			this.message = message;
		}

		public override long Decode(byte[] obj)
		{
			scale = 1;
			long modid = BitConverter.ToInt64(obj, 0);
			id = BitConverter.ToInt64(obj, sizeof(long) * 1);
			ttl = BitConverter.ToInt32(obj, sizeof(long) * 2);
			double vx = BitConverter.ToDouble(obj, sizeof(long) * 2 + sizeof(int) * 1);
			double vy = BitConverter.ToDouble(obj, sizeof(long) * 2 + sizeof(int) * 1 + sizeof(double) * 1);
			Origin = new Vector2D(vx, vy);
			int index = sizeof(long) * 2 + sizeof(int) * 1 + sizeof(double) * 2;
			message = Encoding.UTF8.GetString(obj, index, obj.Length - index);
			
			return modid;
		}



		public override void Write()
		{
			try
			{

				Vector2D origin = Origin;
				if (MyAPIGateway.Session?.Camera == null) return;
				Regex ColorEncoder = new Regex("<color=(\\w*?|\\d*?\\,\\d*?\\,\\d*?|\\d*?,\\d*?,\\d*?,\\d*?)>", RegexOptions.IgnoreCase);
				Color textcolor = Color.White;
				var matches = ColorEncoder.Split(message);
				double aspectratio = (double)MyAPIGateway.Session.Config.ScreenWidth / (double)MyAPIGateway.Session.Config.ScreenHeight;
				var inv_aspectratio = 1 / aspectratio;
				var fov = MyAPIGateway.Session.Camera.FovWithZoom;
				double localscale = 0.1735 * Math.Tan(fov / 2);//exact :)

				origin.X *= localscale * (aspectratio * 9d / 16d);//close enough
				origin.Y *= localscale * inv_aspectratio * (aspectratio * 9d / 16d);
			
				float size = 0.002f;
				float plusx = 0;
				float plusy = 0;
				int i = 0;
				foreach (var match in matches)
				{
					i++;
					if (i % 2 == 0)
					{
						Color foundColor;
						if (FontTexture.TryGetColor(match, out foundColor))
						{
							textcolor = foundColor;
						}
						continue;
					}
					foreach (var letter in match)
					{
						if (letter == '\n')
						{
							plusy -= (0.0025f * (float)(localscale * 9d));
							plusx = 0;
							continue;
						}
						size = WriteLetter(origin + new Vector2D(plusx, plusy), letter, textcolor, localscale * 9d);
						plusx += size;
					}
				}

			}
			catch
			{

			}


		}
		private float WriteLetter(Vector2D Origin, char letter, Color fontcolor, double scale)
		{
			FontTexture.FontMatrix value;

			MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X, Origin.Y, -0.1), headmatrix);
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor.ToVector4(), textpos, headmatrix.Up, headmatrix.Left, 0.002f * (float)scale);
				return ((float)value.ex / (float)45) * 0.0025f * (float)scale;
			}
			return 0.002f * (float)scale;
		}
	}
}
