﻿using Sandbox.ModAPI;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using VRage.Game;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messages
{
	class SpaceMessage : MessageBase
	{
		public Vector3D pos;
		public Vector3D up;
		public Vector3D left;
		public FontTexture.TextOrientation orient;
		public override long Decode(byte[] obj)
		{
			try
			{
				int index = 0;
				ushort modver = BitConverter.ToUInt16(obj, index);
				ushort msgtype = BitConverter.ToUInt16(obj, index += sizeof(short));
				if (msgtype != 1) return -1;//fail
				long modid = BitConverter.ToInt64(obj, index +=  sizeof(short));
				id = BitConverter.ToInt64(obj, index += sizeof(long));
				ttl = BitConverter.ToInt32(obj, index += sizeof(long));
				orient = (FontTexture.TextOrientation)obj[index += sizeof(int)];
				pos = new Vector3D(BitConverter.ToDouble(obj, index += 1), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				
				up = new Vector3D(BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				left = new Vector3D(BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)), BitConverter.ToDouble(obj, index += sizeof(double)));
				up.Normalize();
				left.Normalize();
				scale = BitConverter.ToDouble(obj, index += sizeof(double));
				index += sizeof(double);
				message = Encoding.UTF8.GetString(obj, index, obj.Length - index);
				//MyAPIGateway.Utilities.ShowMessage("recieved", index.ToString());
				return modid;
			}
			catch
			{
				return -1;//fail
			}
			
		}

		public override void Write()
		{

			if (Vector3.DistanceSquared(MyAPIGateway.Session.Camera.WorldMatrix.Translation, pos) > Math.Pow( 200d * (scale < 1.0d ? 1.0d : scale), 2))
				return;
			Regex ColorEncoder = new Regex("<color=(\\w*?|\\d*?\\,\\d*?\\,\\d*?|\\d*?,\\d*?,\\d*?,\\d*?)>", RegexOptions.IgnoreCase);
			Vector4 textcolor = Color.White.ToVector4();
			var matches = ColorEncoder.Split(message);
			float size = 0;
			float xstart = 0;
			if (orient == FontTexture.TextOrientation.center || orient == FontTexture.TextOrientation.rtl)
				xstart = -calculateStart(matches);
			if(orient == FontTexture.TextOrientation.rtl)
			{
				matches.Reverse();
			}
			if(orient == FontTexture.TextOrientation.center)
			{
				xstart = xstart / 2f;
			}
			float plusx = xstart;

			float plusy = 0;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					Color foundColor;
					if (FontTexture.TryGetColor(match, out foundColor))
					{
						textcolor = foundColor.ToVector4();
					}
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						plusy -= (float)(scale) * 1.2f;
						plusx = xstart;

						continue;
					}
					Vector3D vec_x = -new Vector3D(left) * plusx;
					Vector3D vec_y = new Vector3D(up) * plusy;
		
					size = WriteLetter(pos + vec_x + vec_y, letter, textcolor, scale);
					plusx += size;
				}
			}
		}

		private float calculateStart(string[] matches)
		{
			float cnt = 0.0f;
			float mxcnt = 0.0f;
			FontTexture.FontMatrix value;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						if(cnt > mxcnt)
							mxcnt = cnt;
						cnt = 0.0f;
						continue;
					}
					if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
						cnt+= ((float)value.ex / (float)45) * (float)scale * 1.2f;
					else
						cnt+= ((float)15 / (float)45) * (float)scale * 1.2f;
				}
			}
			if (cnt > mxcnt)
				mxcnt = cnt;
			return mxcnt;
		}

	

		private float WriteLetter(Vector3D textpos, char letter, Vector4 fontcolor, double scale)
		{
			FontTexture.FontMatrix value;
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(letter != ' ')//skip spaces save on draw calls
					VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor, textpos, up, left, (float)scale);
				return ((float)value.ex / (float)45) * (float)scale * 1.2f;
			}
			return ((float)15 / (float)45) * (float)scale * 1.2f;
		}
	}
}
