﻿using Sandbox.ModAPI;
using System;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messagesv2
{
	class ModBillboardHUDMessage : ModMessageBase
	{

		private enum EntityMembers : int
		{
			Message = 0,
			Origin = 10,
			Options,
			BillBoardColor,
			Material,
			Rotation,
			Width,
			Height
		}
		public Vector2D Origin = Vector2D.Zero;
		public MyStringId Material = MyStringId.GetOrCompute("");
		public Options Options = Options.None;
		public Color BillBoardColor = Color.White;
		public float Rotation = 0f;
		public float Width = 1f;
		public float Height = 1f;

		public override object GetMember(int MemberEnum)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.Origin:
					return Origin;
				case EntityMembers.Options:
					return (byte)Options;
				case EntityMembers.BillBoardColor:
					return BillBoardColor;
				case EntityMembers.Material:
					return Material;
				case EntityMembers.Rotation:
					return Rotation;

				case EntityMembers.Message:
					return Material;
				case EntityMembers.Width:
					return Width;
				case EntityMembers.Height:
					return Height;

				default:
					return base.GetMember(MemberEnum);
			}
		}
		public override void SetMember(int MemberEnum, object Value)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.Origin:
					Origin = (Vector2D)Value;
					return;
				case EntityMembers.Options:
					Options = (Options)Value;
					return;
				case EntityMembers.BillBoardColor:
					BillBoardColor = (Color)Value;
					return;
				case EntityMembers.Material:
					Material = (MyStringId)Value;
					//Message = Message.Clear().Append(Material.ToString());
					return;
				case EntityMembers.Rotation:
					Rotation = (float)Value;
					return;

				case EntityMembers.Message:
					// do not allow message to be set. 
					//Message = (StringBuilder)Value;
					//Material = MyStringId.GetOrCompute(Message.ToString());
					return;

				case EntityMembers.Width:
					Width = Math.Abs((float)Value);
					return;
				case EntityMembers.Height:
					Height = Math.Abs((float)Value);
					return;

				default:
					base.SetMember(MemberEnum, Value);
					return;
			}
		}

		public override void Draw()
		{

			try
			{
				if (Options.HasFlag(Options.HideHud))
				{
					if (FontTexture.hudHidden)
						return;//hud is hidden dont draw. 
				}
				if (MyAPIGateway.Session?.Camera == null) return;
				Vector2D origin = Origin;
				var fov = MyAPIGateway.Session.Camera.FovWithZoom;
				double aspectratio = MyAPIGateway.Session.Camera.ViewportSize.X / MyAPIGateway.Session.Camera.ViewportSize.Y;
				double localscale = 0.1 * Math.Tan(fov / 2);
				origin += Offset;
				origin.X *= (localscale * aspectratio);
				origin.Y *= localscale;

				float w = Width * (float)Scale;
                float h = Height * (float)Scale;
                if (!Options.HasFlag(Options.Fixed))
				{
					w *= (float)(localscale * aspectratio);
					h *= (float)localscale;
				}

				WriteBillboard(origin, Material, BillBoardColor, w / 2f, h / 2f, (float)localscale);
			}
			
			catch
			{

			}
		}
		private void WriteBillboard(Vector2D Origin, MyStringId billboard, Vector4 billboardcolor, float w, float h, float multi)
		{
			MatrixD cameramatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X, Origin.Y, (Options.HasFlag(Options.Shadowing) ? -0.10001 : -0.1)), cameramatrix);
			var rotquad = Quaternion.CreateFromAxisAngle(cameramatrix.Forward, Rotation);
			cameramatrix = MatrixD.Transform(cameramatrix, rotquad);
			VersionHelper.AddBillboardOriented(billboard, billboardcolor, textpos, cameramatrix.Up, cameramatrix.Left,   w ,  h, Blend);
			return;
		}
	}
}
