﻿using Sandbox.ModAPI;
using System;
using VRage.ModAPI;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messagesv2
{


	class ModEntityMessage : ModMessageBase
	{
		enum EntityMembers : int
		{
			Entity = 10,
			LocalPosition,
			Up,
			Forward,
			Orientation,
			Max,
			TransformMatrix
		}

		IMyEntity Entity;
		Vector3D rel = Vector3D.Zero;
		Vector3D up = MatrixD.Identity.Up;
		Vector3D forward = MatrixD.Identity.Forward;
		MatrixD local = MatrixD.Identity;
		FontTexture.TextOrientation Orient;
		Vector2D Max;
		Color textcolor = Color.White;

		public override void Close()
		{
			Entity = null;
		}

		void UpdateMatrix()
		{
			local = MatrixD.CreateWorld(rel, forward, up);
		}
		void UpdateForwardUp()
		{
			forward = local.Forward;
			up = local.Up;
			rel = local.Translation;
		}
		void UpdateForward(Vector3D Val)
		{
			if(forward != Val)
			{
				forward = Val;
				UpdateMatrix();
			}
		}
		void UpdateUp(Vector3D Val)
		{
			if (up != Val)
			{
				up = Val;
				UpdateMatrix();
			}
		}
		void UpdateLocalPosition(Vector3D Val)
		{
			if(Val != rel)
			{
				rel = Val;
				UpdateMatrix();
			}
		}

		private float calculateStart(string[] matches)
		{
			float cnt = 0.0f;
			float mxcnt = 0.0f;
			FontTexture.FontMatrix value;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						if (cnt > mxcnt)
							mxcnt = cnt;
						cnt = 0.0f;
						continue;
					}
					if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
					{
						cnt += ((float)value.ex / (float)45) * (float)Scale * 1.2f;
					}
					else
						cnt += ((float)15 / (float)45) * (float)Scale * 1.2f;

				}
			}
			if (cnt > mxcnt)
				mxcnt = cnt;

			return mxcnt;
		}

		public override object GetMember(int MemberEnum)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.Entity:
					return Entity;
				case EntityMembers.Forward:
					return local.Forward;
				case EntityMembers.Up:
					return local.Up;
				case EntityMembers.Max:
					return Max;
				case EntityMembers.LocalPosition:
					return local.Translation;
				case EntityMembers.Orientation:
					return (int)Orient;
				case EntityMembers.TransformMatrix:
					return local;

				default:
					return base.GetMember(MemberEnum);
			}
		}
		public override void SetMember(int MemberEnum, object Value)
		{
			switch((EntityMembers)MemberEnum)
			{
				case EntityMembers.Entity:
					Entity = (IMyEntity)Value;
					return;
				case EntityMembers.Forward:
					UpdateForward((Vector3D)Value);
					return;
				case EntityMembers.Up:
					UpdateUp((Vector3D)Value);
					return;
				case EntityMembers.Max:
					Max = (Vector2D)Value;
					return;
				case EntityMembers.LocalPosition:
					UpdateLocalPosition((Vector3D)Value);
                    return;
				case EntityMembers.Orientation:
					Orient = (FontTexture.TextOrientation)(byte)Value;
					return;
				case EntityMembers.TransformMatrix:
					local = (MatrixD)Value;
					UpdateForwardUp();
					return;
				default:
                    base.SetMember(MemberEnum, Value);
					return;
			}
			
		}
		public override Vector2D GetLength()
		{
			Vector2D Length = Vector2D.Zero;
			var matches = ColorEncoder.Split(Message.ToString());

			var fov = MyAPIGateway.Session.Camera.FovWithZoom;
			double localscale = 0.1d * Math.Tan(fov / 2d);
			double lx = 0d;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					//skip
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						Length.Y -= (0.0025f * (float)(localscale * 9d * Scale));
						if (lx > Length.X) Length.X = lx;
						Length.X = 0;
						continue;
					}
					lx += WriteLetter(MatrixD.Identity, letter, Color.Black, localscale * 9d * Scale, false);
				}
			}
			if (lx > Length.X) Length.X = lx;
			return Length;
		}
		public override void Draw()
		{
			if (Entity == null)
			{
				MarkForClose();
				return;
			}
			try
			{
				
                if (Vector3.DistanceSquared(MyAPIGateway.Session.Camera.WorldMatrix.Translation, Entity.WorldMatrix.Translation) > Math.Pow(200d * (Scale < 1.0d ? 1.0d : Scale), 2))
					return;
				MatrixD current = Entity.WorldMatrix;
				var rot = Quaternion.CreateFromRotationMatrix(local.GetOrientation());
				var rot2 = Quaternion.CreateFromRotationMatrix(current.GetOrientation());
				var rotby = Quaternion.Concatenate(rot, rot2);
				current = MatrixD.CreateFromQuaternion(rotby);
				Vector3D Translation = Entity.WorldMatrix.Translation + Vector3D.Transform(local.Translation, current);
				var matches = ColorEncoder.Split(Message.ToString());



				float size = 0;
				float xstart = 0;
				if (Orient == FontTexture.TextOrientation.center || Orient == FontTexture.TextOrientation.rtl)
					xstart = -calculateStart(matches);
				if (Orient == FontTexture.TextOrientation.center)
				{
					xstart = xstart / 2f;
				}
				double plusx = xstart + Offset.X;
				double plusy = 0 + Offset.Y;
				int i = 0;
				foreach (var match in matches)
				{
					i++;
					if (i % 2 == 0)
					{
						Color foundColor;
						if (FontTexture.TryGetColor(match, out foundColor))
						{
							textcolor = foundColor;
						}
						continue;
					}
					foreach (var letter in match)
					{
						if (letter == '\n')
						{
							plusy -= (Scale) * 1.2d;
							plusx = xstart;
							continue;
						}

						Vector3D vec_x = current.Left * -plusx;
						Vector3D vec_y = current.Up * plusy;
						current.Translation = Translation + vec_x + vec_y;

						if ((plusx < Max.X || Max.X <= 0) && (plusy > -Max.Y || Max.Y <= 0))
						{
							size = WriteLetter(current, letter, textcolor, Scale);
							plusx += size;
						}

					}
				}

			}
			catch
			{
				return;
			}
			base.Draw();
		}
		private float WriteLetter(MatrixD world, char letter, Color fontcolor, double scale, bool draw = true)
		{
			FontTexture.FontMatrix value;
			
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(draw)
				{
					VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor.ToVector4(), world.Translation, world.Up, world.Left, (float)scale, Blend);
				}
				return ((float)value.ex / (float)45) * (float)scale * 1.2f;
			}
			return ((float)15 / (float)45) * (float)scale * 1.2f;
		}

	}
}
