﻿using Sandbox.ModAPI;
using System;
using VRage.Utils;
using VRageMath;

namespace UIFun.Messagesv2
{
	class ModHUDMessage : ModMessageBase
	{

		private enum EntityMembers : int
		{
			Origin = 10,
			Options,
			ShadowColor,
		}
		public Vector2D Origin = Vector2D.Zero;
		public Options Options = Options.None;
		public Color ShadowColor = Color.Black;

		public override object GetMember(int MemberEnum)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.Origin:
					return Origin;
				case EntityMembers.Options:
					return (byte)Options;
				case EntityMembers.ShadowColor:
					return ShadowColor;

				default:
					return base.GetMember(MemberEnum);
			}
		}
		public override void SetMember(int MemberEnum, object Value)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.Origin:
					Origin = (Vector2D)Value;
					return;
				case EntityMembers.Options:
					Options = (Options)Value;
					return;
				case EntityMembers.ShadowColor:
					ShadowColor = (Color)Value;
					return;
				default:
					base.SetMember(MemberEnum, Value);
					return;
			}
		}

		public override Vector2D GetLength()
		{
			Vector2D Length = Vector2D.Zero;
			var matches = ColorEncoder.Split(Message.ToString());

			var fov = MyAPIGateway.Session.Camera.FovWithZoom;
			double localscale = 0.1d * Math.Tan(fov / 2d);
			double aspectratio = MyAPIGateway.Session.Camera.ViewportSize.X / MyAPIGateway.Session.Camera.ViewportSize.Y;
			double lx = 0d;
			int i = 0;
			bool eol = false;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					//skip
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\r')
						continue;
					if (letter == '\n')
					{
						Length.Y -= 0.025 * localscale * Scale;
						if (lx > Length.X) Length.X = lx;
						lx = 0;
						eol = true;
						continue;
					}
					eol = false;
					lx += WriteLetter(Vector2D.Zero, letter, Color.Black, localscale * Scale, false);
				}
			}
			if(!eol)
			{
				Length.Y -= 0.025 * localscale * Scale;
			}
			if (lx > Length.X) Length.X = lx;
			lx *= (1d / aspectratio);
            MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
			Vector3D textpos = Vector3D.Transform(new Vector3D(Length.X, Length.Y, -0.1), headmatrix);
			var ScreenPos = MyAPIGateway.Session.Camera.WorldToScreen(ref textpos);

			return new Vector2D(ScreenPos.X, ScreenPos.Y);
		}
		public override void Draw()
		{

			if (Options.HasFlag(Options.HideHud))
			{
				if (FontTexture.hudHidden)
					return;//hud is hidden dont draw. 
			}
			Vector2D origin = Origin;
			if (MyAPIGateway.Session?.Camera == null) return;
				
			Vector4 textcolor = Color.White.ToVector4();
			var matches = ColorEncoder.Split(Message.ToString());

			var fov = MyAPIGateway.Session.Camera.FovWithZoom;
			double aspectratio = MyAPIGateway.Session.Camera.ViewportSize.X / MyAPIGateway.Session.Camera.ViewportSize.Y;
			double localscale = 0.1 * Math.Tan(fov / 2);
			origin = origin + Offset;
			origin.X *= localscale * (aspectratio );
			origin.Y *= localscale;

			
			double plusx = 0d;
			double plusy = 0d;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					Color foundColor;
					if (FontTexture.TryGetColor(match, out foundColor))
					{
						textcolor = foundColor.ToVector4();
					}
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\r')
						continue;
					if (letter == '\n')
					{
						plusy -= (0.025d * (localscale * Scale));
						plusx = 0d;
						continue;
					}
					if (Options.HasFlag(Options.Shadowing))
						WriteShadow(origin + new Vector2D(plusx, plusy), letter, ShadowColor, localscale *  Scale);
					plusx += WriteLetter(origin + new Vector2D(plusx, plusy), letter, textcolor, localscale *  Scale) ;
				}
			}
		}


		private double WriteLetter(Vector2D Origin, char letter, Vector4 fontcolor, double scale, bool draw = true)
		{
            FontTexture.FontMatrix value;
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(draw)
				{
					MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
					Vector3D textpos = Vector3D.Transform(new Vector3D((Origin.X) + ( 1 - (value.ex / 90d ) ) * 0.025f * scale, Origin.Y - (0.0125f * scale), -0.1), headmatrix);
					if (letter != ' ')
						VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor, textpos, headmatrix.Up, headmatrix.Left, 0.02f * (float)scale, Blend);
				}
				return (value.ex / 45d)  * scale * 0.025d;
			}
			return 0.33d * scale * 0.025d;
		}
		private void WriteShadow(Vector2D Origin, char letter, Vector4 fontcolor, double scale)
		{

			FontTexture.FontMatrix value;


			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				MatrixD headmatrix = MyAPIGateway.Session.Camera.WorldMatrix;
				Vector3D textpos = Vector3D.Transform(new Vector3D(Origin.X + (value.ex / 90d) * 0.025f * scale, Origin.Y - (0.0175f * scale), -0.10001), headmatrix);
				if (letter != ' ')
					VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute("EngineThrustMiddle"), fontcolor, textpos, headmatrix.Up, headmatrix.Left, 0.03f * (float)scale, Blend);
				//return ((float)value.ex / (float)45)  * (float)scale;
			}
			//return 0.33f * (float)scale;
		}
	}
}
