﻿using Sandbox.ModAPI;
using System;
using System.Linq;
using VRage.Utils;
using VRageMath;


namespace UIFun.Messagesv2
{

	public enum TextOrientation : byte
	{
		ltr = 1,
		center = 2,
		rtl = 3
	}
	class ModSpaceMessage : ModMessageBase
	{
		private enum EntityMembers : int
		{
			WorldPosition = 10,
			Up,
			Left,
			TxtOrientation

		}
		public Vector3D WorldPosition = Vector3D.Zero;
		public Vector3D Up = MatrixD.Identity.Up;
		public Vector3D Left = MatrixD.Identity.Left;
		public TextOrientation TxtOrientation = TextOrientation.ltr;

		public override object GetMember(int MemberEnum)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.WorldPosition:
					return WorldPosition;
				case EntityMembers.Up:
					return Up;
				case EntityMembers.Left:
					return Left;
				case EntityMembers.TxtOrientation:
					return (int)TxtOrientation;

				default:
					return base.GetMember(MemberEnum);
			}
		}
		public override void SetMember(int MemberEnum, object Value)
		{
			switch ((EntityMembers)MemberEnum)
			{
				case EntityMembers.WorldPosition:
					WorldPosition = (Vector3D)Value;
					return;
				case EntityMembers.Up:
					Up = (Vector3D)Value;
					return;
				case EntityMembers.Left:
					Left = (Vector3D)Value;
					return;
				case EntityMembers.TxtOrientation:
					TxtOrientation = (TextOrientation)(byte)Value;
					return;
				default:
					base.SetMember(MemberEnum, Value);
					return;
			}

		}
		public override Vector2D GetLength()
		{
			Vector2D Length = Vector2D.Zero;
			var matches = ColorEncoder.Split(Message.ToString());

			var fov = MyAPIGateway.Session.Camera.FovWithZoom;
			double localscale = 0.1d * Math.Tan(fov / 2d);
			double lx = 0d;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					//skip
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						Length.Y -= (0.0025f * (float)(localscale * 9d * Scale));
						if (lx > Length.X) Length.X = lx;
						Length.X = 0;
						continue;
					}
					lx += WriteLetter(Vector3D.Zero, letter, Color.Black, localscale * 9d * Scale, false);
				}
			}
			if (lx > Length.X) Length.X = lx;
			return Length;
		}
		public override void Draw()
		{

			if (Vector3.DistanceSquared(MyAPIGateway.Session.Camera.WorldMatrix.Translation, WorldPosition) > Math.Pow(200d * (Scale < 1.0d ? 1.0d : Scale), 2))
				return;
			
			Vector4 textcolor = Color.White.ToVector4();
			var matches = ColorEncoder.Split(Message.ToString());
			float size = 0;
			float xstart = 0;
			if (TxtOrientation == TextOrientation.center || TxtOrientation == TextOrientation.rtl)
				xstart = -calculateStart(matches);
			if (TxtOrientation == TextOrientation.rtl)
			{
				matches.Reverse();
			}
			if (TxtOrientation == TextOrientation.center)
			{
				xstart = xstart / 2f;
			}
			double plusx = xstart + Offset.X;

			double plusy = Offset.Y;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					Color foundColor;
					if (FontTexture.TryGetColor(match, out foundColor))
					{
						textcolor = foundColor.ToVector4();
					}
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						plusy -= (float)(Scale) * 1.2f;
						plusx = xstart;

						continue;
					}
					Vector3D vec_x = -new Vector3D(Left) * plusx;
					Vector3D vec_y = new Vector3D(Up) * plusy;

					size = WriteLetter(WorldPosition + vec_x + vec_y, letter, textcolor, Scale);
					plusx += size;
				}
			}
		}

		private float calculateStart(string[] matches)
		{
			float cnt = 0.0f;
			float mxcnt = 0.0f;
			FontTexture.FontMatrix value;
			int i = 0;
			foreach (var match in matches)
			{
				i++;
				if (i % 2 == 0)
				{
					continue;
				}
				foreach (var letter in match)
				{
					if (letter == '\n')
					{
						if (cnt > mxcnt)
							mxcnt = cnt;
						cnt = 0.0f;
						continue;
					}
					if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
						cnt += ((float)value.ex / (float)45) * (float)Scale * 1.2f;
					else
						cnt += ((float)15 / (float)45) * (float)Scale * 1.2f;
				}
			}
			if (cnt > mxcnt)
				mxcnt = cnt;
			return mxcnt;
		}



		private float WriteLetter(Vector3D textpos, char letter, Vector4 fontcolor, double scale, bool draw = true)
		{
			FontTexture.FontMatrix value;
			if (FontTexture.instance.FontDict.TryGetValue(letter, out value))
			{
				if(draw)
				{
					if (letter != ' ')//skip spaces save on draw calls
						VersionHelper.AddBillboardOriented(MyStringId.GetOrCompute(value.material), fontcolor, textpos, Up, Left, (float)scale, Blend);
				}
				return ((float)value.ex / (float)45) * (float)scale * 1.2f;
			}
			return ((float)15 / (float)45) * (float)scale * 1.2f;
		}
	}
}
